"""
    请求中间件
"""
from typing import Union
from loguru import logger
from palp.network.request import Request
from palp.network.response import Response
from palp.spider.spider_base import BaseSpider


class BaseRequestMiddleware:
    def request_in(self, spider: BaseSpider, request: Request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """
        pass

    def request_close(self, spider: BaseSpider, request: Request, response: Response) -> Union[Request, None]:
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """
        return

    def request_error(self, spider: BaseSpider, request: Request, exception_type: str, exception: str) -> Union[
        Request, None]:
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return: [Request, None]
        """

        logger.error(exception)

        return

    def request_failed(self, spider: BaseSpider, request: Request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """
        pass


# 用于外部引用，避免写类型
class RequestMiddleware:
    def request_in(self, spider, request) -> None:
        """
        请求进入时的操作

        :param spider:
        :param request:
        :return:
        """
        pass

    def request_close(self, spider, request, response):
        """
        请求结束时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param response:
        :return: [Request, None]
        """
        return

    def request_error(self, spider, request, exception_type: str, exception: str):
        """
        请求出错时的操作

        :param spider:
        :param request: 该参数可返回（用于放弃当前请求，并发起新请求）
        :param exception_type: 错误的类型
        :param exception: 错误的详细信息
        :return: [Request, None]
        """

        logger.error(exception)

        return

    def request_failed(self, spider, request) -> None:
        """
        超过最大重试次数时的操作

        :param spider:
        :param request:
        :return:
        """
        pass



"""
    redis set 过滤

    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
    print(BloomFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
"""
from palp import settings
from palp.network.request import Request
from palp.filter.filter_base import BaseFilter
from palp.conn.conn_redis import Redis, RedisLock


class RequestRedisFilter(BaseFilter):

    def is_repeat(self, obj, **kwargs) -> bool:
        """
        获取对应的指纹，通过 redis 的 set 去重

        :param obj:
        :param kwargs:
        :return:
        """
        fingerprint_md5 = self.fingerprint(obj=obj)

        if isinstance(obj, Request):
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_REQUEST
        else:
            redis_key_filter = settings.REDIS_KEY_QUEUE_FILTER_ITEM

        with RedisLock(lock_name=settings.REDIS_KEY_LOCK):
            if Redis.conn().sismember(redis_key_filter, fingerprint_md5):
                return True

            Redis.conn().sadd(redis_key_filter, fingerprint_md5)

            return False



"""
    创建 palp 项目
"""
import re
import shutil
from pathlib import Path


class CreateProject:
    def __init__(self, project_name: str):
        """

        :param project_name: 项目名字
        """
        self.project_name = project_name
        self.path = Path('.').absolute()  # 相对路径
        self.path_template = Path(__file__).absolute().parent.parent.joinpath('template').joinpath('project')  # 模板路径

    def create(self):
        """
        复制项目文件

        :return:
        """
        self.check_name()

        parent_path = self.path.joinpath(self.project_name.title())

        shutil.copytree(self.path_template, parent_path)

    def check_name(self):
        """
        检查名字是否符合要求

        :return:
        """
        matcher = re.match(r'[0-9a-zA-Z_]+', self.project_name)

        if not re.match(r'[0-9_]', self.project_name) and matcher and len(matcher.group()) == len(self.project_name):
            return True

        raise NameError('Palp 项目名字请以字母开头，并仅含有数字字母下划线！')


if __name__ == '__main__':
    CreateProject(project_name='demo').create()



"""
    shell 交互生成文件

    测试 python shell.py xxx
"""
import sys
import argparse
from palp.tool.create_spider import CreateSpider
from palp.tool.create_project import CreateProject


def generator():
    if len(sys.argv) == 1:
        sys.argv.append('--h')  # 无参数默认输出帮助

    parser = argparse.ArgumentParser(description="Palp 创建爬虫模板")
    parser.add_argument('-p', '--project', help='创建爬虫项目如：palp create -p <project_name>')
    parser.add_argument('-s', '--spider', nargs='+', help='创建爬虫如：palp create -s <spider_name> <spider_type>')
    args = parser.parse_args()

    # 创建 spider
    if args.spider:
        args_spider = args.spider
        if len(args_spider) == 1:
            spider_name = args_spider[0]
            spider_type = 1
        else:
            spider_name = args_spider[0]
            spider_type = int(args_spider[1])

        if spider_type not in [1, 2]:
            raise ValueError('spider 只能是 1 或 2 ！')
        CreateSpider(spider_name, spider_type).create()

    # 创建项目
    if args.project:
        project_name = args.project
        CreateProject(project_name).create()


def helper():
    print("palp 操作命令如下")
    print("\tpalp create [options] [args]")
    print("\n可选 options:")
    cmds = {"-p": "即 project，创建爬虫项目", "-s": "即 spider，创建爬虫"}
    for cmdname, cmdclass in sorted(cmds.items()):
        print("\t%s\t\t%s" % (cmdname, cmdclass))

    print("\n可选 args:")
    cmds = {"1": "创建 spider 时，创建普通 spider", "2": "创建 spider 时，创建分布式 spider"}
    for cmdname, cmdclass in sorted(cmds.items()):
        print("\t%s\t\t%s" % (cmdname, cmdclass))

    print("\n示例:")
    print("\tpalp create -p demo")
    print("\tpalp create -s baidu 1")


def main():
    args = sys.argv
    if len(args) < 2:
        helper()
        return

    command = args.pop(1)
    if command == "create":
        generator()
    else:
        helper()


if __name__ == '__main__':
    helper()



"""
    请求的处理
"""
import urllib3
import requests
from palp import settings
from typing import Callable
from palp.tool.user_agent import random_ua
from requests.cookies import RequestsCookieJar
from palp.network.response import ResponseDownloader
from urllib3.exceptions import InsecureRequestWarning

# 禁用不安全警告
urllib3.disable_warnings(InsecureRequestWarning)


class Request:
    # requests 模块所需的
    REQUEST_ATTRS = [
        'url',
        'method',
        'params',
        'data',
        'headers',
        'cookies',
        'files',
        'auth',
        'timeout',
        'allow_redirects',
        'proxies',
        'hooks',
        'stream',
        'verify',
        'cert',
        'json',
    ]

    # 框架运行所需的
    PALP_ATTRS = [
        'filter_repeat',
        'keep_session',
        'callback',
        'render',
        'render_time',
        'session',
        'cookie_jar'
    ]

    def __init__(self, url: str = None, method: str = None, params=None, data=None, headers=None, cookies=None,
                 files=None, auth=None, timeout=None, allow_redirects=True, proxies=None, hooks=None, stream=None,
                 verify=None, cert=None, json=None, filter_repeat: bool = None, keep_session: bool = None,
                 callback: Callable = None, render: bool = None, render_time: int = None, level: int = None,
                 session: requests.Session = None, cookie_jar: RequestsCookieJar = None, **kwargs):
        """
        requests 参数
        :param url: 请求链接
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: 文件：{ file_name:open(file) }
        :param auth: 身份验证
        :param timeout: 超时时间
        :param allow_redirects: 允许重定向：默认 True
        :param proxies: 代理：{'http':'http://ip:port','https':'https://ip:port'}
        :param hooks: 钩子：{'response':func}
        :param stream: 流传输
        :param verify: 证书验证：默认 False
        :param cert: ssl 证书路径
        :param json: json：也可以使用 data = json.dumps(data)

        Palp 参数
        :param filter_repeat: 是否过滤请求
        :param keep_session: 是否保持 session
        :param callback: 回调函数
        :param render: 是否用浏览器渲染
        :param render_time: 等待页面加载最大时间
        :param level: 启用优先级队列时的优先级，分数越大，优先级约低

        Palp 参数（非用户设置）
        :param session: session 保证每个线程都是独立的
        :param cookie_jar: cookie_jar，存储 cookie

        传递的参数
        :param kwargs: 其余需要传递的参数，若参数名不存在则返回 None
        """

        # Request 所需字段
        self._requests_params = {}  # request 参数
        self._original_params = {}  # __init__ 后的 request 参数，用来传递快速发起请求
        self.keep_session = keep_session or False
        self.filter_repeat = filter_repeat or False
        self.callback = callback
        self.render = render or False
        self.render_time = render_time or 60
        self.session = session or requests.session()
        self.level = level or 10
        self.cookie_jar = cookie_jar or RequestsCookieJar()

        # requests 请求参数
        self.url = url
        self.method = method
        self.params = params
        self.data = data
        self.headers = headers or {}
        self.cookies = cookies
        self.files = files
        self.auth = auth
        self.timeout = timeout or settings.REQUEST_TIMEOUT or 60
        self.allow_redirects = allow_redirects
        if not proxies and settings.REQUEST_PROXIES_TUNNEL_URL:
            proxies = {'http': settings.REQUEST_PROXIES_TUNNEL_URL, 'https': settings.REQUEST_PROXIES_TUNNEL_URL}
        self.proxies = proxies
        self.hooks = hooks
        self.stream = stream
        self.verify = verify
        self.cert = cert
        self.json = json

        # 传递参数更新一下
        for key, value in kwargs.items():
            self.__dict__[key] = value

        self._original_params.update(kwargs)

        # 判断类型
        if self.method:
            self.method = self.method.upper()
        elif self.data or self.json:
            self.method = 'POST'
        else:
            self.method = 'GET'

        # 添加 ua
        if settings.RANDOM_USERAGENT:
            ua = random_ua()
        else:
            ua = settings.DEFAULT_USER_AGENT

        if not self.headers:
            self.headers.update({'User-Agent': ua})
        elif self.headers.get('User-Agent'):
            self.headers.update({'User-Agent': ua})
        elif self.headers.get('user-agent'):
            self.headers.update({'user-agent': ua})

    def send(self):
        """

        :return: ResponseParser 的解析器
        """
        response = ResponseDownloader(
            keep_session=self.keep_session,
            session=self.session,
            cookie_jar=self.cookie_jar
        ).response(**self._requests_params)

        # 给响应体携带上请求原数据（非纯原）
        response.__dict__.update({
            'palp': {
                'requests': self._original_params
            }
        })

        return self.callback, response

    def __setattr__(self, key, value):
        """
        实现 requests.xxx 设置参数（只有 self.xxx 才会进入该函数）

        @param key:
        @param value:
        @return:
        """
        self.__dict__[key] = value

        if key in self.__class__.REQUEST_ATTRS:
            self._requests_params[key] = value

        if not key.startswith('_'):
            self._original_params[key] = value

    def __str__(self):
        return f"<Request {self.method}-{self.url}>"



"""
    分布式 spider
"""
import time
from palp import settings
from threading import Thread
from abc import abstractmethod
from palp.buffer.buffer_item import ItemBuffer
from palp.spider.spider_base import BaseSpider
from palp.tool.client_heart import ClientHeart
from palp.conn.conn_redis import RedisLockNoWait, Redis


class DistributiveSpider(BaseSpider, Thread):
    def __init__(self, thread_count=None, redis_key: str = None, request_filter=False, item_filter=False):
        """

        :param thread_count: 线程数量
        :param redis_key: 自定义 redis_key 的名字
        :param request_filter: 开启请求过滤，不然 filter_repeat 无效
        :param item_filter: 开启 item 过滤
        """
        setattr(settings, 'SPIDER_TYPE', 2)
        self.redis_key = redis_key or self.name

        # 根据 spider 的名字设置前缀
        setattr(settings, 'REDIS_KEY_QUEUE_ITEM', settings.REDIS_KEY_QUEUE_ITEM.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_MASTER', settings.REDIS_KEY_MASTER.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_STOP', settings.REDIS_KEY_STOP.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_LOCK', settings.REDIS_KEY_LOCK.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_HEARTBEAT', settings.REDIS_KEY_HEARTBEAT.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_QUEUE_REQUEST', settings.REDIS_KEY_QUEUE_REQUEST.format(redis_key=self.redis_key))

        setattr(settings, 'REDIS_KEY_HEARTBEAT_FAILED',
                settings.REDIS_KEY_HEARTBEAT_FAILED.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_QUEUE_FILTER_REQUEST',
                settings.REDIS_KEY_QUEUE_FILTER_REQUEST.format(redis_key=self.redis_key))
        setattr(settings, 'REDIS_KEY_QUEUE_FILTER_ITEM',
                settings.REDIS_KEY_QUEUE_FILTER_ITEM.format(redis_key=self.redis_key))

        super().__init__(thread_count, request_filter, item_filter)

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    def parse(self, request, response) -> None:
        """
        解析

        :param request:
        :param response:
        :return:
        """
        pass

    def competition_for_master(self) -> None:
        """
        竞争为 master

        :return:
        """
        if Redis.conn().exists(settings.REDIS_KEY_MASTER):
            return

        with RedisLockNoWait(lock_name=settings.REDIS_KEY_LOCK) as lock:
            if lock.lock_success:
                self.spider_master = True
                Redis.conn().set(settings.REDIS_KEY_MASTER, '')

    def spider_logic(self):
        """
        spider 处理逻辑

        :return:
        """
        # master 竞争
        self.competition_for_master()

        # master 机器处理的事情
        if self.spider_master:
            # 删除停止标志
            Redis.conn().delete(settings.REDIS_KEY_STOP)
            # 清空心跳
            Redis.conn().delete(settings.REDIS_KEY_HEARTBEAT, settings.REDIS_KEY_HEARTBEAT_FAILED)

            # 处理 item
            ItemBuffer.from_settings()
            self._item_buffer = ItemBuffer(spider=self, q=self._queue_item)
            self._item_buffer.start()

            # 分发任务
            self.distribute_task()
        else:
            time.sleep(1)  # 睡 1s 避免停止标志没被移除

        # 等待所有任务执行结束
        ClientHeart(spider=self).start()

        # master 机器处理的事情
        if self.spider_master:
            # 删除 master 的标志
            Redis.conn().delete(settings.REDIS_KEY_MASTER)
            # 删除 stop 标志
            ClientHeart.remove_stop_status()
            # 判断是否移除 filter
            if not settings.PERSISTENCE_REQUEST_FILTER:
                Redis.conn().delete(settings.REDIS_KEY_QUEUE_FILTER_REQUEST)
            if not settings.PERSISTENCE_ITEM_FILTER:
                Redis.conn().delete(settings.REDIS_KEY_QUEUE_FILTER_ITEM)



"""
    item 去重

    setting 启用才会去重，但是会提前存指纹
"""
from palp import settings
from palp.filter.filter_bloom import BloomFilter
from palp.pipeline.pipeline_base import BasePipeline
from palp.filter.filter_redis import RequestRedisFilter
from palp.filter.filter_memory import RequestMemoryFilter
from palp.exception.exception_drop import DropItemException
from palp.filter.filter_redis_bloom import RequestRedisBloomFilter


# 本地：基于 python set 的去重
class ItemMemoryFilterPipeline(BasePipeline):
    def __init__(self):
        self.item_memory_filter = RequestMemoryFilter()

    def pipeline_in(self, spider, item) -> None:
        if settings.ITEM_FILTER:
            is_repeat = self.item_memory_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


# 本地：基于 bloom 的去重
class ItemBloomFilterPipeline(BasePipeline):
    def __init__(self):
        self.item_bloom_filter = BloomFilter()

    def pipeline_in(self, spider, item) -> None:
        if settings.ITEM_FILTER:
            is_repeat = self.item_bloom_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


# redis：基于 redis set 的去重
class ItemRedisFilterPipeline(BasePipeline):
    def __init__(self):
        self.item_redis_filter = RequestRedisFilter()

    def pipeline_in(self, spider, item) -> None:
        if settings.ITEM_FILTER:
            is_repeat = self.item_redis_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


# redis：基于 redis 的 bloom 去重
class ItemRedisBloomFilterPipeline(BasePipeline):
    def __init__(self):
        self.item_redis_bloom_filter = RequestRedisBloomFilter()

    def pipeline_in(self, spider, item) -> None:
        if settings.ITEM_FILTER:
            is_repeat = self.item_redis_bloom_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")



from palp.tool import start_spider
from palp.item.item_base import Item
from palp.network.request import Request
from palp.network.response import Response
from palp.pipeline.pipeline_base import Pipeline
from palp.spider import Spider, DistributiveSpider
from palp.network.request_method import RequestGet, RequestPost
from palp.conn.conn_redis import Redis, RedisLock, RedisLockNoWait
from palp.middleware.middleware_spider_base import SpiderMiddleware
from palp.middleware.middleware_request_base import RequestMiddleware
from palp.exception.exception_drop import DropItemException, DropRequestException



from palp.pipeline.pipeline_item_filter import ItemBloomFilterPipeline, ItemRedisBloomFilterPipeline, \
    ItemRedisFilterPipeline, ItemMemoryFilterPipeline

from palp.pipeline.pipeline_base import Pipeline



__all__ = [
    "DropRequestException",
    "DropItemException",
    "NotGeneratorFunctionError",
]

from palp.exception.exception_error import NotGeneratorFunctionError
from palp.exception.exception_drop import DropRequestException, DropItemException



"""
    本地内存去重

    print(MemoryFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
    print(MemoryFilter.is_repeat(request=RequestGet(url='https://www.baidu.com')))
"""
from palp.network.request import Request
from palp.filter.filter_base import BaseFilter, FilterLock


class RequestMemoryFilter(BaseFilter):
    def __init__(self):
        self.memory_filter_request = set()
        self.memory_filter_item = set()

    def is_repeat(self, obj, **kwargs) -> bool:
        """
        获取对应的指纹，通过 python 的 set 去重

        :param obj:
        :param kwargs:
        :return:
        """
        fingerprint_md5 = self.fingerprint(obj=obj)

        if isinstance(obj, Request):
            memory_filter = self.memory_filter_request
        else:
            memory_filter = self.memory_filter_item

        with FilterLock():
            if fingerprint_md5 in memory_filter:
                return True

            memory_filter.add(fingerprint_md5)

            return False



"""
    以进程方式启动指定数量个爬虫实例
"""
import importlib
from pathlib import Path
from loguru import logger
from multiprocessing import Process
from palp.spider.spider_base import BaseSpider


# 启动指定 spider
class SpiderRunner(Process):
    def __init__(self, spider: BaseSpider, count: int = 1, **kwargs):
        """

        :param spider:
        :param count: 要运行的数量
        :param kwargs:
        """
        super().__init__()
        self.spider = spider
        self.count = count
        self.kwargs = kwargs

    def run(self) -> None:
        for _ in range(self.count):
            Process(target=self.runner, args=(self.spider, self.kwargs)).start()

    @staticmethod
    def runner(*args):
        spider = args[0]
        spider(**args[1]).start()


def load_spider() -> dict:
    """
    获取所有 spider 实例

    :return:
    """
    path = Path('.').absolute().parent

    # 获取到 spider 的目录
    spider_dir = None
    for i in path.iterdir():
        if i.is_dir() and i.name.lower().startswith('spider'):
            spider_dir = i
            break

    # 获取所有 spider
    spider_modules = {}
    if spider_dir:
        for spider_file in spider_dir.iterdir():
            if spider_file.name.startswith('_'):
                continue
            spider_path = f"{spider_dir.name}.{spider_file.name.split('.')[0]}"
            spider_module = importlib.import_module(spider_path)

            # 获取 spider 的信息
            for key, value in spider_module.__dict__.items():
                if key.startswith('_'):
                    continue
                elif value.__dict__.get('__module__') == spider_path:
                    spider_modules[value.__dict__.get('spider_name')] = value
                    break
    else:
        logger.warning('未找到 spider 文件夹，请将该文件放在与 settings 同级目录')

    return spider_modules


def execute(spider_name: str, count: int = 1, **kwargs):
    """
    启动爬虫进程

    :param spider_name: 爬虫名字
    :param count: 启动几个实例
    :param kwargs: 爬虫启动参数
    :return:
    """
    spider_modules = load_spider()

    if spider_name not in spider_modules:
        logger.error(f'无当前 spider，请检查 {spider_name} 是否存在')
        return

    SpiderRunner(spider=spider_modules[spider_name], count=count, **kwargs).start()
